/* $Id: pathfind.cpp 18943 2007-07-21 07:28:04Z mordante $ */
/*
Copyright (C) 2003 by David White <dave@whitevine.net>
Copyright (C) 2005 - 2007 by Guillaume Melquiond <guillaume.melquiond@gmail.com>
Part of the Battle for Wesnoth Project http://www.wesnoth.org/

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License version 2.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY.

See the COPYING file for more details.
*/

#include "global.hpp"

#include "astarnode.hpp"
#include "gamestatus.hpp"
#include "log.hpp"
#include "pathfind.hpp"
#include "util.hpp"
#include "wassert.hpp"

class gamestatus;

#include <cmath>
#include <iostream>

#define LOG_PF LOG_STREAM(info, engine)

static gamemap::location find_vacant(const gamemap& map,
		const unit_map& units,
		const gamemap::location& loc, int depth,
		VACANT_TILE_TYPE vacancy,
		std::set<gamemap::location>& touched)
	{
		if(touched.count(loc))
			return gamemap::location();

		touched.insert(loc);

		if (map.on_board(loc) && units.find(loc) == units.end() &&
		    (vacancy == VACANT_ANY || map.is_castle(loc))) {
			return loc;
		} else if(depth == 0) {
			return gamemap::location();
		} else {
			gamemap::location adj[6];
			get_adjacent_tiles(loc,adj);
			for(int i = 0; i != 6; ++i) {
				if(!map.on_board(adj[i]) || vacancy == VACANT_CASTLE && !map.is_castle(adj[i]))
					continue;

				const gamemap::location res =
					find_vacant(map, units, adj[i], depth - 1, vacancy, touched);

				if (map.on_board(res))
					return res;
			}

			return gamemap::location();
		}
	}

gamemap::location find_vacant_tile(const gamemap& map,
																	 const unit_map& units,
																	 const gamemap::location& loc,
																	 VACANT_TILE_TYPE vacancy)
{
	for(int i = 1; i != 50; ++i) {
		std::set<gamemap::location> touch;
		const gamemap::location res = find_vacant(map,units,loc,i,vacancy,touch);
		if(map.on_board(res))
			return res;
	}

	return gamemap::location();
}

bool enemy_zoc(gamemap const &map,
               unit_map const &units,
               std::vector<team> const &teams,
               gamemap::location const &loc, team const &viewing_team, unsigned int side, bool see_all)
{
	gamemap::location locs[6];
	const team &current_team = teams[side-1];
	get_adjacent_tiles(loc,locs);
	for(int i = 0; i != 6; ++i) {
	  unit_map::const_iterator it;	     
	  it = find_visible_unit(units, locs[i], map, teams, viewing_team,see_all);

	  if (it != units.end() && it->second.side() != side &&
		 current_team.is_enemy(it->second.side()) && it->second.emits_zoc()) {
	    return true;
	  }
	}

	return false;
}

static void find_routes(const gamemap& map, const gamestatus& status,
		const game_data& gamedata,
		const unit_map& units,
		const unit& u,
		const gamemap::location& loc,
		int move_left,
		std::map<gamemap::location,paths::route>& routes,
		std::vector<team> const &teams,
		bool force_ignore_zocs, bool allow_teleport, int turns_left,
		  bool starting_pos, const team &viewing_team, bool see_all)
	{
		if(size_t(u.side()-1) >= teams.size()) {
			return;
		}

		team const &current_team = teams[u.side()-1];

		//find adjacent tiles
		std::vector<gamemap::location> locs(6);
		get_adjacent_tiles(loc,&locs[0]);

		//check for teleporting units -- we must be on a vacant (or occupied by this unit)
		//village, that is controlled by our team to be able to teleport.
		if (allow_teleport && map.is_village(loc) &&
		    current_team.owns_village(loc) &&
			(starting_pos || find_visible_unit(units, loc, map,
										teams, viewing_team,see_all) == units.end())) {
			const std::vector<gamemap::location>& villages = map.villages();

			//if we are on a village, see all friendly villages that we can
			//teleport to
			for(std::vector<gamemap::location>::const_iterator t = villages.begin();
			    t != villages.end(); ++t) {
				if (!current_team.owns_village(*t) || units.count(*t))
					continue;

				locs.push_back(*t);
			}
		}

		//iterate over all adjacent tiles
		for(size_t i = 0; i != locs.size(); ++i) {
			const gamemap::location& currentloc = locs[i];

			//check if the adjacent location is off the board
			if (currentloc.x < 0 || currentloc.y < 0 ||
			    currentloc.x >= map.x() || currentloc.y >= map.y())
				continue;

			//see if the tile is on top of an enemy unit
			const unit_map::const_iterator unit_it =
			  find_visible_unit(units, locs[i], map, teams, viewing_team,see_all);

			if (unit_it != units.end() &&
			    current_team.is_enemy(unit_it->second.side()))
				continue;

			//find the terrain of the adjacent location
			const t_translation::t_letter terrain = map[currentloc.x][currentloc.y];

			//find the movement cost of this type onto the terrain
			const int move_cost = u.movement_cost(terrain);
			if (move_cost <= move_left ||
			    turns_left > 0 && move_cost <= u.total_movement()) {
				int new_move_left = move_left - move_cost;
				int new_turns_left = turns_left;
				if (new_move_left < 0) {
					--new_turns_left;
					new_move_left = u.total_movement() - move_cost;
				}
				const bool skirmisher = force_ignore_zocs | u.get_ability_bool("skirmisher",currentloc);
				const bool zoc = !skirmisher && enemy_zoc(map,units,teams,currentloc, viewing_team,u.side(),see_all);

				const int zoc_move_left = zoc ? 0 : new_move_left;
				const int total_movement = new_turns_left * u.total_movement() + zoc_move_left;

				const std::map<gamemap::location,paths::route>::const_iterator
					rtit = routes.find(currentloc);

				//if a better route to that tile has already been found
				if(rtit != routes.end() && rtit->second.move_left >= total_movement)
					continue;
				paths::route new_route = routes[loc];
				new_route.steps.push_back(loc);

				new_route.move_left = u.total_movement() * new_turns_left + zoc_move_left;
				routes[currentloc] = new_route;

				if (new_route.move_left > 0) {
					find_routes(map, status, gamedata, units, u, currentloc,
					            zoc_move_left, routes, teams, force_ignore_zocs,
					            allow_teleport, new_turns_left, false, viewing_team, see_all);
				}
			}
		}
	}

paths::paths(gamemap const &map, gamestatus const &status,
             game_data const &gamedata,
             unit_map const &units,
             gamemap::location const &loc,
             std::vector<team> const &teams,
	     bool force_ignore_zoc,
             bool allow_teleport, const team &viewing_team,
		   int additional_turns, bool see_all)
{
	const unit_map::const_iterator i = units.find(loc);
	if(i == units.end()) {
		std::cerr << "unit not found\n";
		return;
	}

	routes[loc].move_left = i->second.movement_left();
	find_routes(map,status,gamedata,units,i->second,loc,
		i->second.movement_left(),routes,teams,force_ignore_zoc,
		allow_teleport,additional_turns,true,viewing_team, see_all);
}

int route_turns_to_complete(const unit &u, const gamemap &map, paths::route &rt, 
		const unit_map &units, const std::vector<team>& teams)
{
	if(rt.steps.empty())
		return 0;

	int turns = 0, movement = u.movement_left();

	for(std::vector<gamemap::location>::const_iterator i = rt.steps.begin()+1;
	    i != rt.steps.end(); ++i) {
		wassert(map.on_board(*i));
		const int move_cost = u.movement_cost(map[i->x][i->y]);
		movement -= move_cost;

		if (movement < 0) {
			++turns;
			rt.turn_waypoints.insert(std::make_pair(*(i-1), turns));
			movement = u.total_movement() - move_cost;
			if(movement < 0) {
				return -1;
			}
		}

		if (!u.get_ability_bool("skirmisher",*i)) {
			gamemap::location adj[6];
			get_adjacent_tiles(*i, adj);

			for (size_t j = 0; j != 6; ++j) {
				unit_map::const_iterator enemy_unit = find_visible_unit(units, adj[j], map, teams, teams[u.side()-1]);
				if (enemy_unit != units.end() && teams[u.side()-1].is_enemy(enemy_unit->second.side())
					&& enemy_unit->second.emits_zoc()) {
					 movement = 0;
				}
			}
		}
	}
	

	//add "end-of-path" to waypoints.
	if (turns > 0) {
		rt.turn_waypoints.insert(std::make_pair(*(rt.steps.end()-1), turns+1));
	} else if (movement==0 || map.is_village(*(rt.steps.end()-1))) {
		rt.turn_waypoints.insert(std::make_pair(*(rt.steps.end()-1), 1));
	} else {
		rt.turn_waypoints.insert(std::make_pair(*(rt.steps.end()-1), 0));
	}
	return turns;
}


shortest_path_calculator::shortest_path_calculator(unit const &u, team const &t, unit_map const &units,
                                                   std::vector<team> const &teams, gamemap const &map)
	: unit_(u), team_(t), units_(units), teams_(teams), map_(map),
	  movement_left_(unit_.movement_left()),
	  total_movement_(unit_.total_movement())
{
}

double shortest_path_calculator::cost(const gamemap::location& /*src*/,const gamemap::location& loc, const double so_far, const bool isDst) const
{
	wassert(map_.on_board(loc));

	//the location is not valid
	//1. if the loc is shrouded, or
	//2. if moving in it costs more than the total movement of the unit, or
	//3. if there is a visible enemy on the hex, or	
	//4. if the unit is not a skirmisher and there is a visible enemy with
	//   a ZoC on an adjacent hex in the middle of the route
	// #4 is a bad criteria!  It should be that moving into a ZOC uses up
	// the rest of your moves

	if (team_.shrouded(loc.x, loc.y))
		return getNoPathValue();

	int const base_cost = unit_.movement_cost(map_[loc.x][loc.y]);
	wassert(base_cost >= 1); // pathfinding heuristic: the cost must be at least 1
	if (total_movement_ < base_cost)
		return getNoPathValue();

	unit_map::const_iterator
		enemy_unit = find_visible_unit(units_, loc, map_, teams_, team_),
		units_end = units_.end();

	if (enemy_unit != units_end && team_.is_enemy(enemy_unit->second.side()))
		return getNoPathValue();

	//compute how many movement points are left in the game turn needed to
	//reach the previous hex
	//total_movement_ is not zero, thanks to the pathfinding heuristic
	int remaining_movement = movement_left_ - static_cast<int>(so_far);
	if (remaining_movement < 0)
		remaining_movement = total_movement_ - (-remaining_movement) % total_movement_;

	//supposing we had 2 movement left, and wanted to move onto a hex which
	//takes 3 movement, it's going to cost us 5 movement in total, since we
	//sacrifice this turn's movement. Take that into account here.
	int additional_cost = base_cost > remaining_movement ? remaining_movement : 0;

	if (!isDst && !unit_.get_ability_bool("skirmisher",loc)) {
	  gamemap::location adj[6];
	  get_adjacent_tiles(loc, adj);
	  
	  for (size_t i = 0; i != 6; ++i) {
	    enemy_unit = find_visible_unit(units_, adj[i], map_, teams_, team_);
	    if (enemy_unit != units_end && team_.is_enemy(enemy_unit->second.side()) &&
		   enemy_unit->second.emits_zoc()){
		 //should cost us remaining movement.
		 //		 return getNoPathValue();
		 return total_movement_ + additional_cost;
	    }
	  }
	}

	return base_cost + additional_cost;
}

emergency_path_calculator::emergency_path_calculator(const unit& u, const gamemap& map)
	: unit_(u), map_(map)
{
}

double emergency_path_calculator::cost(const gamemap::location&,const gamemap::location& loc, const double, const bool) const
{
	wassert(map_.on_board(loc));

	return unit_.movement_cost(map_[loc.x][loc.y]);
}
